/**
 * Copyright (c) NVIDIA CORPORATION & AFFILIATES, 2021. ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */

#ifndef UCM_OPTS_H_
#define UCM_OPTS_H_

#include <ucs/sys/compiler_def.h>

BEGIN_C_DECLS

void ucs_init_ucm_opts();
void ucs_cleanup_ucm_opts();

END_C_DECLS

#endif
